@@ -1,4 +1,4 @@
-# Copyright 2010-2019 The pygit2 contributors
+# Copyright 2010-2021 The pygit2 contributors
 #
 # This file is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2,
@@ -28,17 +28,19 @@ from io import BytesIO
 from string import hexdigits
 import tarfile
 from time import time
+import warnings
 
 # Import from pygit2
-from _pygit2 import Repository as _Repository, init_file_backend
-from _pygit2 import Oid, GIT_OID_HEXSZ, GIT_OID_MINPREFIXLEN
-from _pygit2 import GIT_CHECKOUT_SAFE, GIT_CHECKOUT_RECREATE_MISSING, GIT_DIFF_NORMAL
-from _pygit2 import GIT_FILEMODE_LINK
-from _pygit2 import GIT_BRANCH_LOCAL, GIT_BRANCH_REMOTE, GIT_BRANCH_ALL
-from _pygit2 import GIT_REF_SYMBOLIC
-from _pygit2 import Reference, Tree, Commit, Blob
-from _pygit2 import InvalidSpecError
+from ._pygit2 import Repository as _Repository, init_file_backend
+from ._pygit2 import Oid, GIT_OID_HEXSZ, GIT_OID_MINPREFIXLEN
+from ._pygit2 import GIT_CHECKOUT_SAFE, GIT_CHECKOUT_RECREATE_MISSING, GIT_DIFF_NORMAL
+from ._pygit2 import GIT_FILEMODE_LINK
+from ._pygit2 import GIT_BRANCH_LOCAL, GIT_BRANCH_REMOTE, GIT_BRANCH_ALL
+from ._pygit2 import GIT_REF_SYMBOLIC
+from ._pygit2 import Reference, Tree, Commit, Blob, Signature
+from ._pygit2 import InvalidSpecError
 
+from .callbacks import git_fetch_options
 from .config import Config
 from .errors import check_error
 from .ffi import ffi, C
@@ -47,11 +49,12 @@ from .remote import RemoteCollection
 from .blame import Blame
 from .utils import to_bytes, StrArray
 from .submodule import Submodule
+from .packbuilder import PackBuilder
 
 
 class BaseRepository(_Repository):
-    def __init__(self, backend, *args, **kwargs):
-        super().__init__(backend, *args, **kwargs)
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
         self._common_init()
 
     def _common_init(self):
@@ -81,10 +84,88 @@ class BaseRepository(_Repository):
         object."""
         return self.odb.write(*args, **kwargs)
 
+    def pack(self, path=None, pack_delegate=None, n_threads=None):
+        """Pack the objects in the odb chosen by the pack_delegate function
+        and write `.pack` and `.idx` files for them.
+
+        Returns: the number of objects written to the pack
+
+        Parameters:
+
+        path
+            The path to which the `.pack` and `.idx` files should be written. `None` will write to the default location.
+
+        pack_delegate
+            The method which will provide add the objects to the pack builder. Defaults to all objects.
+
+        n_threads
+            The number of threads the `PackBuilder` will spawn. If set to 0, libgit2 will autodetect the number of CPUs.
+        """
+
+        def pack_all_objects(pack_builder):
+            for obj in self.odb:
+                pack_builder.add(obj)
+
+        pack_delegate = pack_delegate or pack_all_objects
+
+        builder = PackBuilder(self)
+        if n_threads is not None:
+            builder.set_threads(n_threads)
+        pack_delegate(builder)
+        builder.write(path=path)
+
+        return builder.written_objects_count
+
+
     def __iter__(self):
         return iter(self.odb)
 
+    def add_submodule(self, url, path, link=True, callbacks=None):
+        """Add a submodule to the index.
+
+        Returns: the submodule that was added.
+
+        Parameters:
+
+        url
+            The URL of the submdoule.
+
+        path
+            The path within the parent repository to add the submodule
+
+        link
+            Should workdir contain a gitlink to the repo in `.git/modules` vs. repo directly in workdir.
+        """
+        csub = ffi.new('git_submodule **')
+        curl = ffi.new('char[]', to_bytes(url))
+        cpath = ffi.new('char[]', to_bytes(path))
+        gitlink = 1 if link else 0
+
+        err = C.git_submodule_add_setup(csub, self._repo, curl, cpath, gitlink)
+        check_error(err)
+
+        submodule_instance = Submodule._from_c(self, csub[0])
+
+        # prepare options
+        opts = ffi.new('git_submodule_update_options *')
+        C.git_submodule_update_init_options(opts, C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION)
+
+        with git_fetch_options(callbacks, opts=opts.fetch_opts) as payload:
+            crepo = ffi.new('git_repository **')
+            err = C.git_submodule_clone(crepo, submodule_instance._subm, opts)
+            payload.check_error(err)
+
+        # clean-up the submodule repository
+        Repository._from_c(crepo[0], True)
+
+        err = C.git_submodule_add_finalize(submodule_instance._subm)
+        check_error(err)
+        return submodule_instance
+
     def lookup_submodule(self, path):
+        """
+        Look up submodule information by name or path.
+        """
         csub = ffi.new('git_submodule **')
         cpath = ffi.new('char[]', to_bytes(path))
 
@@ -93,28 +174,26 @@ class BaseRepository(_Repository):
         return Submodule._from_c(self, csub[0])
 
     def update_submodules(self, submodules=None, init=False, callbacks=None):
+        """
+        Update a submodule. This will clone a missing submodule and checkout
+        the subrepository to the commit specified in the index of the
+        containing repository. If the submodule repository doesn't contain the
+        target commit (e.g. because fetchRecurseSubmodules isn't set), then the
+        submodule is fetched using the fetch options supplied in options.
+        """
         if submodules is None:
             submodules = self.listall_submodules()
 
         # prepare options
         opts = ffi.new('git_submodule_update_options *')
-        C.git_submodule_update_init_options(
-            opts,
-            C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION)
-        if callbacks is not None:
-            callbacks._fill_fetch_options(opts.fetch_opts)
-        if init:
-            i = 1
-        else:
-            i = 0
+        C.git_submodule_update_init_options(opts, C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION)
 
-        for submodule in submodules:
-            submodule_instance = self.lookup_submodule(submodule)
-            err = C.git_submodule_update(
-                submodule_instance._subm,
-                i,
-                opts)
-            check_error(err)
+        with git_fetch_options(callbacks, opts=opts.fetch_opts) as payload:
+            i = 1 if init else 0
+            for submodule in submodules:
+                submodule_instance = self.lookup_submodule(submodule)
+                err = C.git_submodule_update(submodule_instance._subm, i, opts)
+                payload.check_error(err)
 
         return None
 
@@ -141,10 +220,7 @@ class BaseRepository(_Repository):
     # Remotes
     #
     def create_remote(self, name, url):
-        """Create a new remote. Return a <Remote> object.
-
-        This method is deprecated, please use Remote.remotes.create()
-        """
+        warnings.warn("Use repo.remotes.create(..)", DeprecationWarning)
         return self.remotes.create(name, url)
 
     #
@@ -180,7 +256,7 @@ class BaseRepository(_Repository):
     #
     # References
     #
-    def create_reference(self, name, target, force=False):
+    def create_reference(self, name, target, force=False, message=None):
         """Create a new reference "name" which points to an object or to
         another reference.
 
@@ -189,10 +265,13 @@ class BaseRepository(_Repository):
 
         Keyword arguments:
 
-        force
+        force: bool
             If True references will be overridden, otherwise (the default) an
             exception is raised.
 
+        message: str
+            Optional message to use for the reflog.
+
         Examples::
 
             repo.create_reference('refs/heads/foo', repo.head.target)
@@ -206,9 +285,11 @@ class BaseRepository(_Repository):
                 and GIT_OID_MINPREFIXLEN <= len(target) <= GIT_OID_HEXSZ))
 
         if direct:
-            return self.create_reference_direct(name, target, force)
+            return self.create_reference_direct(name, target, force,
+                                                message=message)
 
-        return self.create_reference_symbolic(name, target, force)
+        return self.create_reference_symbolic(name, target, force,
+                                              message=message)
 
     def resolve_refish(self, refish):
         """Convert a reference-like short name "ref-ish" to a valid
@@ -272,13 +353,13 @@ class BaseRepository(_Repository):
         copts, refs = Repository._checkout_args_to_options(**kwargs)
         check_error(C.git_checkout_head(self._repo, copts))
 
-    def checkout_index(self, **kwargs):
-        """Checkout the repository's index
+    def checkout_index(self, index=None, **kwargs):
+        """Checkout the given index or the repository's index
 
         For arguments, see Repository.checkout().
         """
         copts, refs = Repository._checkout_args_to_options(**kwargs)
-        check_error(C.git_checkout_index(self._repo, ffi.NULL, copts))
+        check_error(C.git_checkout_index(self._repo, index._index if index else ffi.NULL, copts))
 
     def checkout_tree(self, treeish, **kwargs):
         """Checkout the given treeish
@@ -379,7 +460,7 @@ class BaseRepository(_Repository):
             return None
 
         # If it's a string, then it has to be valid revspec
-        if isinstance(obj, str):
+        if isinstance(obj, str) or isinstance(obj, bytes):
             obj = self.revparse_single(obj)
         elif isinstance(obj, Oid):
             obj = self[obj]
@@ -422,19 +503,21 @@ class BaseRepository(_Repository):
             Otherwise the referred object is compared to 'a'
 
         cached
-            if 'b' is None, by default the working directory is compared to 'a'.
+            If 'b' is None, by default the working directory is compared to 'a'.
             If 'cached' is set to True, the index/staging area is used for comparing.
 
         flag
-            a GIT_DIFF_* constant
+            A combination of GIT_DIFF_* constants. For a list of the constants,
+            with a description, see git_diff_option_t in
+            https://github.com/libgit2/libgit2/blob/master/include/git2/diff.h
 
         context_lines
-            the number of unchanged lines that define the boundary
-            of a hunk (and to display before and after)
+            The number of unchanged lines that define the boundary of a hunk
+            (and to display before and after)
 
         interhunk_lines
-            the maximum number of unchanged lines between hunk
-            boundaries before the hunks will be merged into a one
+            The maximum number of unchanged lines between hunk boundaries
+            before the hunks will be merged into a one
 
         Examples::
 
@@ -531,6 +614,8 @@ class BaseRepository(_Repository):
 
         options = ffi.new('git_blame_options *')
         C.git_blame_init_options(options, C.GIT_BLAME_OPTIONS_VERSION)
+        if flags:
+            options.flags = flags
         if min_match_characters:
             options.min_match_characters = min_match_characters
         if newest_commit:
@@ -560,39 +645,88 @@ class BaseRepository(_Repository):
         """Index representing the repository's index file."""
         cindex = ffi.new('git_index **')
         err = C.git_repository_index(cindex, self._repo)
-        check_error(err, True)
+        check_error(err, io=True)
 
         return Index.from_c(self, cindex)
 
     #
     # Merging
     #
+    _FAVOR_TO_ENUM = {
+        'normal': C.GIT_MERGE_FILE_FAVOR_NORMAL,
+        'ours': C.GIT_MERGE_FILE_FAVOR_OURS,
+        'theirs': C.GIT_MERGE_FILE_FAVOR_THEIRS,
+        'union': C.GIT_MERGE_FILE_FAVOR_UNION,
+    }
 
-    @staticmethod
-    def _merge_options(favor):
+    _MERGE_FLAG_TO_ENUM = {
+        'find_renames': C.GIT_MERGE_FIND_RENAMES,
+        'fail_on_conflict': C.GIT_MERGE_FAIL_ON_CONFLICT,
+        'skip_reuc': C.GIT_MERGE_SKIP_REUC,
+        'no_recursive': C.GIT_MERGE_NO_RECURSIVE,
+    }
+
+    _MERGE_FLAG_DEFAULTS = {
+        'find_renames': True,
+    }
+
+    _MERGE_FILE_FLAG_TO_ENUM = {
+        'standard_style': C.GIT_MERGE_FILE_STYLE_MERGE,
+        'diff3_style': C.GIT_MERGE_FILE_STYLE_DIFF3,
+        'simplify_alnum': C.GIT_MERGE_FILE_SIMPLIFY_ALNUM,
+        'ignore_whitespace': C.GIT_MERGE_FILE_IGNORE_WHITESPACE,
+        'ignore_whitespace_change': C.GIT_MERGE_FILE_IGNORE_WHITESPACE_CHANGE,
+        'ignore_whitespace_eol': C.GIT_MERGE_FILE_IGNORE_WHITESPACE_EOL,
+        'patience': C.GIT_MERGE_FILE_DIFF_PATIENCE,
+        'minimal': C.GIT_MERGE_FILE_DIFF_MINIMAL,
+    }
+
+    _MERGE_FILE_FLAG_DEFAULTS = {}
+
+    @classmethod
+    def _flag_dict_to_bitmask(cls, flag_dict, flag_defaults, mapping, label):
+        """
+        Converts a dict eg {"find_renames": True, "skip_reuc": True} to
+        a bitmask eg C.GIT_MERGE_FIND_RENAMES | C.GIT_MERGE_SKIP_REUC.
+        """
+        merged_dict = {**flag_defaults, **flag_dict}
+        bitmask = 0
+        for k, v in merged_dict.items():
+            enum = mapping.get(k, None)
+            if enum is None:
+                raise ValueError("unknown %s: %s" % (label, k))
+            if v:
+                bitmask |= enum
+        return bitmask
+
+    @classmethod
+    def _merge_options(cls, favor='normal', flags={}, file_flags={}):
         """Return a 'git_merge_opts *'"""
 
-        def favor_to_enum(favor):
-            if favor == 'normal':
-                return C.GIT_MERGE_FILE_FAVOR_NORMAL
-            elif favor == 'ours':
-                return C.GIT_MERGE_FILE_FAVOR_OURS
-            elif favor == 'theirs':
-                return C.GIT_MERGE_FILE_FAVOR_THEIRS
-            elif favor == 'union':
-                return C.GIT_MERGE_FILE_FAVOR_UNION
-            else:
-                return None
-
-        favor_val = favor_to_enum(favor)
+        favor_val = cls._FAVOR_TO_ENUM.get(favor, None)
         if favor_val is None:
-            raise ValueError("unkown favor value %s" % favor)
+            raise ValueError("unknown favor: %s" % favor)
+
+        flags_bitmask = Repository._flag_dict_to_bitmask(
+            flags,
+            cls._MERGE_FLAG_DEFAULTS,
+            cls._MERGE_FLAG_TO_ENUM,
+            "merge flag"
+        )
+        file_flags_bitmask = cls._flag_dict_to_bitmask(
+            file_flags,
+            cls._MERGE_FILE_FLAG_DEFAULTS,
+            cls._MERGE_FILE_FLAG_TO_ENUM,
+            "merge file_flag"
+        )
 
         opts = ffi.new('git_merge_options *')
         err = C.git_merge_init_options(opts, C.GIT_MERGE_OPTIONS_VERSION)
         check_error(err)
 
         opts.file_favor = favor_val
+        opts.flags = flags_bitmask
+        opts.file_flags = file_flags_bitmask
 
         return opts
 
@@ -629,7 +763,7 @@ class BaseRepository(_Repository):
 
         return ret
 
-    def merge_commits(self, ours, theirs, favor='normal'):
+    def merge_commits(self, ours, theirs, favor='normal', flags={}, file_flags={}):
         """
         Merge two arbitrary commits.
 
@@ -653,8 +787,34 @@ class BaseRepository(_Repository):
 
             For all but NORMAL, the index will not record a conflict.
 
-        Both "ours" and "theirs" can be any object which peels to a commit or the id
-        (string or Oid) of an object which peels to a commit.
+        flags
+            A dict of str: bool to turn on or off functionality while merging.
+            If a key is not present, the default will be used. The keys are:
+
+            * find_renames. Detect file renames. Defaults to True.
+            * fail_on_conflict. If a conflict occurs, exit immediately instead
+              of attempting to continue resolving conflicts.
+            * skip_reuc. Do not write the REUC extension on the generated index.
+            * no_recursive. If the commits being merged have multiple merge
+              bases, do not build a recursive merge base (by merging the
+              multiple merge bases), instead simply use the first base.
+
+        file_flags
+            A dict of str: bool to turn on or off functionality while merging.
+            If a key is not present, the default will be used. The keys are:
+
+            * standard_style. Create standard conflicted merge files.
+            * diff3_style. Create diff3-style file.
+            * simplify_alnum. Condense non-alphanumeric regions for simplified
+              diff file.
+            * ignore_whitespace. Ignore all whitespace.
+            * ignore_whitespace_change. Ignore changes in amount of whitespace.
+            * ignore_whitespace_eol. Ignore whitespace at end of line.
+            * patience. Use the "patience diff" algorithm
+            * minimal. Take extra time to find minimal diff
+
+        Both "ours" and "theirs" can be any object which peels to a commit or
+        the id (string or Oid) of an object which peels to a commit.
         """
 
         ours_ptr = ffi.new('git_commit **')
@@ -669,7 +829,7 @@ class BaseRepository(_Repository):
         ours = ours.peel(Commit)
         theirs = theirs.peel(Commit)
 
-        opts = self._merge_options(favor)
+        opts = self._merge_options(favor, flags, file_flags)
 
         ffi.buffer(ours_ptr)[:] = ours._pointer[:]
         ffi.buffer(theirs_ptr)[:] = theirs._pointer[:]
@@ -679,7 +839,7 @@ class BaseRepository(_Repository):
 
         return Index.from_c(self, cindex)
 
-    def merge_trees(self, ancestor, ours, theirs, favor='normal'):
+    def merge_trees(self, ancestor, ours, theirs, favor='normal', flags={}, file_flags={}):
         """
         Merge two trees.
 
@@ -705,6 +865,32 @@ class BaseRepository(_Repository):
             * union. Unique lines from each side will be used.
 
             For all but NORMAL, the index will not record a conflict.
+
+        flags
+            A dict of str: bool to turn on or off functionality while merging.
+            If a key is not present, the default will be used. The keys are:
+
+            * find_renames. Detect file renames. Defaults to True.
+            * fail_on_conflict. If a conflict occurs, exit immediately instead
+              of attempting to continue resolving conflicts.
+            * skip_reuc. Do not write the REUC extension on the generated index.
+            * no_recursive. If the commits being merged have multiple merge
+              bases, do not build a recursive merge base (by merging the
+              multiple merge bases), instead simply use the first base.
+
+        file_flags
+            A dict of str: bool to turn on or off functionality while merging.
+            If a key is not present, the default will be used. The keys are:
+
+            * standard_style. Create standard conflicted merge files.
+            * diff3_style. Create diff3-style file.
+            * simplify_alnum. Condense non-alphanumeric regions for simplified
+              diff file.
+            * ignore_whitespace. Ignore all whitespace.
+            * ignore_whitespace_change. Ignore changes in amount of whitespace.
+            * ignore_whitespace_eol. Ignore whitespace at end of line.
+            * patience. Use the "patience diff" algorithm
+            * minimal. Take extra time to find minimal diff
         """
 
         ancestor_ptr = ffi.new('git_tree **')
@@ -723,7 +909,7 @@ class BaseRepository(_Repository):
         ours = ours.peel(Tree)
         theirs = theirs.peel(Tree)
 
-        opts = self._merge_options(favor)
+        opts = self._merge_options(favor, flags, file_flags)
 
         ffi.buffer(ancestor_ptr)[:] = ancestor._pointer[:]
         ffi.buffer(ours_ptr)[:] = ours._pointer[:]
@@ -1047,8 +1233,8 @@ class BaseRepository(_Repository):
         Calculate how many different commits are in the non-common parts of the
         history between the two given ids.
 
-        Ahead is how many commits are in the ancestry of the 'local' commit
-        which are not in the 'upstream' commit. Behind is the opposite.
+        Ahead is how many commits are in the ancestry of the `local` commit
+        which are not in the `upstream` commit. Behind is the opposite.
 
         Returns: a tuple of two integers with the number of commits ahead and
         behind respectively.
@@ -1084,7 +1270,7 @@ class BaseRepository(_Repository):
         """
         Retrieve an attribute for a file by path.
 
-        Returns: a boolean, None if the value is unspecified, or string with
+        Returns: a boolean, `None` if the value is unspecified, or string with
         the value of the attribute.
 
         Parameters:
@@ -1097,8 +1283,17 @@ class BaseRepository(_Repository):
             The name of the attribute to look up.
 
         flags
-            A combination of GIT_ATTR_CHECK_ flags which determine the
+            A combination of `GIT_ATTR_CHECK_` flags which determine the
             lookup order.
+
+        Examples::
+
+            >>> print(repo.get_attr('splash.bmp', 'binary'))
+            True
+            >>> print(repo.get_attr('splash.bmp', 'unknown-attr'))
+            None
+            >>> repo.get_attr('test.h', 'whitespace')
+            'tab-in-indent,trailing-space'
         """
 
         cvalue = ffi.new('char **')
@@ -1132,7 +1327,7 @@ class BaseRepository(_Repository):
         return (ffi.string(cname).decode('utf-8'), ffi.string(cemail).decode('utf-8'))
 
     def set_ident(self, name, email):
-        """Set the identity to be used for reference operations
+        """Set the identity to be used for reference operations.
 
         Updates to some references also append data to their
         reflog. You can use this method to set what identity will be
@@ -1144,7 +1339,7 @@ class BaseRepository(_Repository):
 
     def revert_commit(self, revert_commit, our_commit, mainline=0):
         """
-        Reverts the given Commit against the given "our" Commit, producing an
+        Revert the given Commit against the given "our" Commit, producing an
         Index that reflects the result of the revert.
 
         Returns: an Index with the result of the revert.
@@ -1178,9 +1373,123 @@ class BaseRepository(_Repository):
 
         return Index.from_c(self, cindex)
 
+    #
+    # Amend commit
+    #
+    def amend_commit(self, commit, refname, author=None,
+                     committer=None, message=None, tree=None,
+                     encoding='UTF-8'):
+        """
+        Amend an existing commit by replacing only explicitly passed values,
+        return the rewritten commit's oid.
 
-class Branches(object):
-    def __init__(self, repository, flag=GIT_BRANCH_ALL, commit=None):
+        This creates a new commit that is exactly the same as the old commit,
+        except that any explicitly passed values will be updated. The new
+        commit has the same parents as the old commit.
+
+        You may omit the `author`, `committer`, `message`, `tree`, and
+        `encoding` parameters, in which case this will use the values
+        from the original `commit`.
+
+        Parameters:
+
+        commit : Commit, Oid, or str
+            The commit to amend.
+
+        refname : Reference or str
+            If not `None`, name of the reference that will be updated to point
+            to the newly rewritten commit. Use "HEAD" to update the HEAD of the
+            current branch and make it point to the rewritten commit.
+            If you want to amend a commit that is not currently the tip of the
+            branch and then rewrite the following commits to reach a ref, pass
+            this as `None` and update the rest of the commit chain and ref
+            separately.
+
+        author : Signature
+            If not None, replace the old commit's author signature with this
+            one.
+
+        committer : Signature
+            If not None, replace the old commit's committer signature with this
+            one.
+
+        message : str
+            If not None, replace the old commit's message with this one.
+
+        tree : Tree, Oid, or str
+            If not None, replace the old commit's tree with this one.
+
+        encoding : str
+            Optional encoding for `message`.
+        """
+
+        # Initialize parameters to pass on to C function git_commit_amend.
+        # Note: the pointers are all initialized to NULL by default.
+        coid = ffi.new('git_oid *')
+        commit_cptr = ffi.new('git_commit **')
+        refname_cstr = ffi.NULL
+        author_cptr = ffi.new('git_signature **')
+        committer_cptr = ffi.new('git_signature **')
+        message_cstr = ffi.NULL
+        encoding_cstr = ffi.NULL
+        tree_cptr = ffi.new('git_tree **')
+
+        # Get commit as pointer to git_commit.
+        if isinstance(commit, (str, Oid)):
+            commit = self[commit]
+        elif isinstance(commit, Commit):
+            pass
+        elif commit is None:
+            raise ValueError("the commit to amend cannot be None")
+        else:
+            raise TypeError("the commit to amend must be a Commit, str, or Oid")
+        commit = commit.peel(Commit)
+        ffi.buffer(commit_cptr)[:] = commit._pointer[:]
+
+        # Get refname as C string.
+        if isinstance(refname, Reference):
+            refname_cstr = ffi.new('char[]', to_bytes(refname.name))
+        elif type(refname) is str:
+            refname_cstr = ffi.new('char[]', to_bytes(refname))
+        elif refname is not None:
+            raise TypeError("refname must be a str or Reference")
+
+        # Get author as pointer to git_signature.
+        if isinstance(author, Signature):
+            ffi.buffer(author_cptr)[:] = author._pointer[:]
+        elif author is not None:
+            raise TypeError("author must be a Signature")
+
+        # Get committer as pointer to git_signature.
+        if isinstance(committer, Signature):
+            ffi.buffer(committer_cptr)[:] = committer._pointer[:]
+        elif committer is not None:
+            raise TypeError("committer must be a Signature")
+
+        # Get message and encoding as C strings.
+        if message is not None:
+            message_cstr = ffi.new('char[]', to_bytes(message, encoding))
+            encoding_cstr = ffi.new('char[]', to_bytes(encoding))
+
+        # Get tree as pointer to git_tree.
+        if tree is not None:
+            if isinstance(tree, (str, Oid)):
+                tree = self[tree]
+            tree = tree.peel(Tree)
+            ffi.buffer(tree_cptr)[:] = tree._pointer[:]
+
+        # Amend the commit.
+        err = C.git_commit_amend(coid, commit_cptr[0], refname_cstr,
+                                 author_cptr[0], committer_cptr[0],
+                                 encoding_cstr, message_cstr, tree_cptr[0])
+        check_error(err)
+
+        return Oid(raw=bytes(ffi.buffer(coid)[:]))
+
+
+class Branches:
+
+    def __init__(self, repository: BaseRepository, flag: int = GIT_BRANCH_ALL, commit=None):
         self._repository = repository
         self._flag = flag
         if commit is not None:
@@ -1194,7 +1503,7 @@ class Branches(object):
             self.local = Branches(repository, flag=GIT_BRANCH_LOCAL, commit=commit)
             self.remote = Branches(repository, flag=GIT_BRANCH_REMOTE, commit=commit)
 
-    def __getitem__(self, name):
+    def __getitem__(self, name: str):
         branch = None
         if self._flag & GIT_BRANCH_LOCAL:
             branch = self._repository.lookup_branch(name, GIT_BRANCH_LOCAL)
@@ -1207,7 +1516,7 @@ class Branches(object):
 
         return branch
 
-    def get(self, key):
+    def get(self, key: str):
         try:
             return self[key]
         except KeyError:
@@ -1218,10 +1527,10 @@ class Branches(object):
             if self._commit is None or self.get(branch_name) is not None:
                 yield branch_name
 
-    def create(self, name, commit, force=False):
+    def create(self, name: str, commit, force=False):
         return self._repository.create_branch(name, commit, force)
 
-    def delete(self, name):
+    def delete(self, name: str):
         self[name].delete()
 
     def _valid(self, branch):
@@ -1242,14 +1551,15 @@ class Branches(object):
         return self.get(name) is not None
 
 
-class References(object):
-    def __init__(self, repository):
+class References:
+
+    def __init__(self, repository: BaseRepository):
         self._repository = repository
 
-    def __getitem__(self, name):
+    def __getitem__(self, name: str):
         return self._repository.lookup_reference(name)
 
-    def get(self, key):
+    def get(self, key: str):
         try:
             return self[key]
         except KeyError:
@@ -1262,10 +1572,10 @@ class References(object):
     def create(self, name, target, force=False):
         return self._repository.create_reference(name, target, force)
 
-    def delete(self, name):
+    def delete(self, name: str):
         self[name].delete()
 
-    def __contains__(self, name):
+    def __contains__(self, name: str):
         return self.get(name) is not None
 
     @property
@@ -1277,12 +1587,40 @@ class References(object):
 
 
 class Repository(BaseRepository):
-    def __init__(self, path, *args, **kwargs):
-        if not isinstance(path, str):
-            path = path.decode('utf-8')
+    def __init__(self, path=None, flags=0):
+        """
+        The Repository constructor will commonly be called with one argument,
+        the path of the repository to open.
 
-        path_backend = init_file_backend(path)
-        super().__init__(backend=path_backend, *args, **kwargs)
+        Alternatively, constructing a repository with no arguments will create
+        a repository with no backends. You can use this path to create
+        repositories with custom backends. Note that most operations on the
+        repository are considered invalid and may lead to undefined behavior if
+        attempted before providing an odb and refdb via set_odb and set_refdb.
+
+        Parameters:
+
+        path : str
+        The path to open - if not provided, the repository will have no backend.
+
+        flags : int
+        Flags controlling how to open the repository can optionally be provided - any combination of:
+        -   GIT_REPOSITORY_OPEN_NO_SEARCH
+        -   GIT_REPOSITORY_OPEN_CROSS_FS
+        -   GIT_REPOSITORY_OPEN_BARE
+        -   GIT_REPOSITORY_OPEN_NO_DOTGIT
+        -   GIT_REPOSITORY_OPEN_FROM_ENV
+        """
+
+        if path is not None:
+            if hasattr(path, "__fspath__"):
+                path = path.__fspath__()
+            if not isinstance(path, str):
+                path = path.decode('utf-8')
+            path_backend = init_file_backend(path, flags)
+            super().__init__(path_backend)
+        else:
+            super().__init__()
 
     @classmethod
     def _from_c(cls, ptr, owned):
